/*
 * Copyright 2010-2024 Australian Signals Directorate
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package au.gov.asd.tac.constellation.views.attributeeditor.editors.operations;

import au.gov.asd.tac.constellation.graph.Graph;
import au.gov.asd.tac.constellation.graph.GraphWriteMethods;
import au.gov.asd.tac.constellation.graph.schema.visual.concept.VisualConcept;
import au.gov.asd.tac.constellation.plugins.PluginException;
import au.gov.asd.tac.constellation.plugins.PluginInfo;
import au.gov.asd.tac.constellation.plugins.PluginInteraction;
import au.gov.asd.tac.constellation.plugins.PluginType;
import au.gov.asd.tac.constellation.plugins.parameters.PluginParameters;
import au.gov.asd.tac.constellation.plugins.templates.PluginTags;
import au.gov.asd.tac.constellation.plugins.templates.SimpleEditPlugin;
import au.gov.asd.tac.constellation.views.attributeeditor.AttributeData;
import java.time.ZoneId;
import java.time.ZonedDateTime;

/**
 * Updates the values of a given date-time attribute for selected vertices, to
 * be in a specific time-zone.
 * <p>
 * Note that this plugin works based on selection
 *
 * @author twilight_sparkle
 */
@PluginInfo(pluginType = PluginType.UPDATE, tags = {PluginTags.MODIFY})
public class UpdateTimeZonePlugin extends SimpleEditPlugin {

    @Override
    public String getName() {
        return "Attribute Editor: Update Time Zone";
    }

    private final AttributeData attributeData;
    private final ZoneId zoneId;

    public UpdateTimeZonePlugin(final ZoneId zoneId, final AttributeData attributeData) {
        this.attributeData = attributeData;
        this.zoneId = zoneId;
    }

    @Override
    protected void edit(final GraphWriteMethods graph, final PluginInteraction interaction, final PluginParameters parameters) throws InterruptedException, PluginException {

        final int attrId = attributeData.getAttributeId();
        switch (attributeData.getElementType()) {
            case VERTEX:
                final int vxSelectedAttribute = VisualConcept.VertexAttribute.SELECTED.get(graph);
                if (vxSelectedAttribute != Graph.NOT_FOUND) {
                    for (int j = 0; j < graph.getVertexCount(); j++) {
                        final int vxId = graph.getVertex(j);
                        final ZonedDateTime dateTime = (ZonedDateTime) graph.getObjectValue(attrId, vxId);
                        if (dateTime != null && graph.getBooleanValue(vxSelectedAttribute, vxId)) {
                            graph.setObjectValue(attrId, vxId, ZonedDateTime.ofInstant(dateTime.toInstant(), zoneId));
                        }
                    }
                }
                break;
            case TRANSACTION:
                final int txSelectedAttribute = VisualConcept.TransactionAttribute.SELECTED.get(graph);
                if (txSelectedAttribute != Graph.NOT_FOUND) {
                    for (int j = 0; j < graph.getTransactionCount(); j++) {
                        final int txId = graph.getTransaction(j);
                        final ZonedDateTime dateTime = (ZonedDateTime) graph.getObjectValue(attrId, txId);
                        if (dateTime != null && graph.getBooleanValue(txSelectedAttribute, txId)) {
                            graph.setObjectValue(attrId, txId, ZonedDateTime.ofInstant(dateTime.toInstant(), zoneId));
                        }
                    }
                }
                break;
            case GRAPH:
                final ZonedDateTime dateTime = (ZonedDateTime) graph.getObjectValue(attrId, 0);
                if (dateTime != null) {
                    graph.setObjectValue(attrId, 0, ZonedDateTime.ofInstant(dateTime.toInstant(), zoneId));
                }
                break;
            default:
                break;
        }
    }
}
