/*
 * Copyright 2010-2024 Australian Signals Directorate
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package au.gov.asd.tac.constellation.graph.attribute;

import au.gov.asd.tac.constellation.utilities.temporal.TemporalConstants;
import au.gov.asd.tac.constellation.utilities.temporal.TemporalFormatting;
import au.gov.asd.tac.constellation.utilities.temporal.TimeZoneUtilities;
import java.time.Instant;
import java.time.LocalDateTime;
import java.time.ZoneOffset;
import java.util.Calendar;
import java.util.Date;
import org.apache.commons.lang3.StringUtils;
import org.openide.util.lookup.ServiceProvider;

/**
 * A description of a local_datetime attribute.
 * <p>
 * Local date-times are considered to be fixed points on a calendar and clock,
 * independent of time-zone. They are represented internally as long primitives.
 * These long primitives give the number of days milliseconds since the epoch to
 * the desired date-time.
 *
 * The object representation of these attribute values are java
 * {@link LocalDateTime} objects.
 *
 * @author sirius
 */
@ServiceProvider(service = AttributeDescription.class)
public final class LocalDateTimeAttributeDescription extends AbstractObjectAttributeDescription<LocalDateTime> {

    public static final String ATTRIBUTE_NAME = "local_datetime";
    public static final Class<LocalDateTime> NATIVE_CLASS = LocalDateTime.class;
    public static final LocalDateTime DEFAULT_VALUE = null;

    public LocalDateTimeAttributeDescription() {
        super(ATTRIBUTE_NAME, NATIVE_CLASS, DEFAULT_VALUE);
    }

    @Override
    protected LocalDateTime convertFromObject(final Object object) {
        try {
            return super.convertFromObject(object);
        } catch (final IllegalArgumentException ex) {
            switch (object) {
                case Date date -> {
                    return LocalDateTime.ofInstant(Instant.ofEpochMilli(date.getTime()), TimeZoneUtilities.UTC);
                }
                case Calendar calendar -> {
                    return LocalDateTime.ofInstant(Instant.ofEpochMilli(calendar.getTimeInMillis()), calendar.getTimeZone().toZoneId());
                }
                case Number number -> {
                    return LocalDateTime.ofInstant(Instant.ofEpochMilli(number.longValue()), TimeZoneUtilities.UTC);
                }
                default -> throw new IllegalArgumentException(String.format(
                        "Error converting Object '%s' to datetime", object.getClass()));
            }
        }
    }

    /**
     * Parse a string in CONSTELLATION's date-time format to a ZonedDateTime.
     *
     * CONSTELLATION's date-time format is that which would be generated by
     * {@link TemporalFormatting#LOCAL_DATE_TIME_FORMATTER}.
     * <p>
     * Parsing isn't strict: the date 2011-99-01 will be accepted and
     * punctuation isn't checked. Since the context is parsing of dates from
     * CONSTELLATION files, this isn't expected to be a problem. However, this
     * should not be taken as an excuse to write syntactically incorrect
     * datetime strings elsewhere.
     * <p>
     * Note that this method directly reads substrings with static indices, as
     * this is significantly faster than using a ZonedDateTimeFormatter.
     *
     * @param string A String in CONSTELLATION's date-time format to be parsed.
     *
     * @return A Calendar representing the input datetime.
     */
    @Override
    protected LocalDateTime convertFromString(final String string) {
        if (StringUtils.isBlank(string)) {
            return getDefault();
        } else {
            try {
                final int year = Integer.parseInt(string.substring(0, 4), 10);
                final int month = Integer.parseInt(string.substring(5, 7), 10);
                final int day = Integer.parseInt(string.substring(8, 10), 10);
                final int hour = Integer.parseInt(string.substring(11, 13), 10);
                final int minute = Integer.parseInt(string.substring(14, 16), 10);
                final int second = Integer.parseInt(string.substring(17, 19), 10);
                final int millisecond = Integer.parseInt(string.substring(20, 23), 10);
                return LocalDateTime.of(year, month, day, hour, minute, second, millisecond
                        * TemporalConstants.NANOSECONDS_IN_MILLISECOND);
            } catch (final StringIndexOutOfBoundsException | NumberFormatException ex) {
                throw new IllegalArgumentException(String.format(
                        "Error converting String '%s' to datetime (expected yyyy-mm-dd hh:MM:ss)", string), ex);
            }
        }

    }

    @Override
    public long getLong(final int id) {
        return data[id] != null ? ((LocalDateTime) data[id]).toEpochSecond(ZoneOffset.UTC) * TemporalConstants.MILLISECONDS_IN_SECOND : 0L;
    }

    @Override
    public void setLong(final int id, final long value) {
        data[id] = LocalDateTime.ofInstant(Instant.ofEpochMilli(value), TimeZoneUtilities.UTC);
    }

    @Override
    public String getString(final int id) {
        return data[id] != DEFAULT_VALUE ? ((LocalDateTime) data[id]).format(TemporalFormatting.LOCAL_DATE_TIME_FORMATTER) : null;
    }

    @Override
    public int hashCode(final int id) {
        return (int) data[id];
    }
}
