/*
 * Copyright 2010-2024 Australian Signals Directorate
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package au.gov.asd.tac.constellation.views.pluginreporter;

import au.gov.asd.tac.constellation.graph.Graph;
import au.gov.asd.tac.constellation.graph.manager.GraphManager;
import au.gov.asd.tac.constellation.plugins.reporting.GraphReportListener;
import au.gov.asd.tac.constellation.plugins.reporting.GraphReportManager;
import au.gov.asd.tac.constellation.plugins.reporting.PluginReport;
import au.gov.asd.tac.constellation.utilities.javafx.JavafxStyleManager;
import au.gov.asd.tac.constellation.views.JavaFxTopComponent;
import au.gov.asd.tac.constellation.views.pluginreporter.panes.PluginReporterPane;
import javafx.application.Platform;
import org.netbeans.api.settings.ConvertAsProperties;
import org.openide.awt.ActionID;
import org.openide.awt.ActionReference;
import org.openide.awt.ActionReferences;
import org.openide.util.NbBundle.Messages;
import org.openide.windows.TopComponent;

/**
 * The PluginReporterTopComponent provides the UI for the entire plugin reporter.
 */
@ConvertAsProperties(
        dtd = "-//au.gov.asd.tac.constellation.views.pluginreporter//PluginReporter//EN",
        autostore = false
)
@TopComponent.Description(
        preferredID = "PluginReporterTopComponent",
        iconBase = "au/gov/asd/tac/constellation/views/pluginreporter/resources/plugin-reporter.png",
        persistenceType = TopComponent.PERSISTENCE_ALWAYS
)
@TopComponent.Registration(
        mode = "rightSlidingSide",
        openAtStartup = true
)
@ActionID(
        category = "Window",
        id = "au.gov.asd.tac.constellation.views.pluginreporter.PluginReporterTopComponent"
)
@ActionReferences({
    @ActionReference(path = "Menu/Views", position = 1000),
    @ActionReference(path = "Shortcuts", name = "CS-P")
})
@TopComponent.OpenActionRegistration(
        displayName = "#CTL_PluginReporterAction",
        preferredID = "PluginReporterTopComponent"
)
@Messages({
    "CTL_PluginReporterAction=Plugin Reporter",
    "CTL_PluginReporterTopComponent=Plugin Reporter",
    "HINT_PluginReporterTopComponent=Plugin Reporter"
})
public final class PluginReporterTopComponent extends JavaFxTopComponent<PluginReporterPane> implements GraphReportListener {

    private final PluginReporterPane reporterPane;

    public PluginReporterTopComponent() {
        initComponents();
        setName(Bundle.CTL_PluginReporterTopComponent());
        setToolTipText(Bundle.HINT_PluginReporterTopComponent());
        reporterPane = new PluginReporterPane();
        initContent();
    }

    /**
     * This method is called from within the constructor to initialize the form. WARNING: Do NOT modify this code. The
     * content of this method is always regenerated by the Form Editor.
     */
    // <editor-fold defaultstate="collapsed" desc="Generated Code">//GEN-BEGIN:initComponents
    private void initComponents() {

        javax.swing.GroupLayout layout = new javax.swing.GroupLayout(this);
        this.setLayout(layout);
        layout.setHorizontalGroup(
            layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGap(0, 400, Short.MAX_VALUE)
        );
        layout.setVerticalGroup(
            layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGap(0, 300, Short.MAX_VALUE)
        );
    }// </editor-fold>//GEN-END:initComponents

    // Variables declaration - do not modify//GEN-BEGIN:variables
    // End of variables declaration//GEN-END:variables
    @Override
    protected void handleComponentOpened() {
        super.handleComponentOpened();
        GraphManager.getDefault().addGraphManagerListener(this);
        GraphReportManager.addGraphReportListener(this);
        handleNewGraph(GraphManager.getDefault().getActiveGraph());
    }

    @Override
    protected void handleComponentClosed() {
        super.handleComponentClosed();
        handleNewGraph(null);
        GraphReportManager.removeGraphReportListener(this);
        GraphManager.getDefault().removeGraphManagerListener(this);
    }

    void writeProperties(java.util.Properties p) {
        // Required for @ConvertAsProperties, intentionally left blank
    }

    void readProperties(java.util.Properties p) {
        // Required for @ConvertAsProperties, intentionally left blank
    }

    @Override
    protected void componentShowing() {
        super.componentShowing();
        handleNewGraph(GraphManager.getDefault().getActiveGraph());
    }

    @Override
    protected void handleNewGraph(final Graph graph) {
        if (needsUpdate()) {
            Platform.runLater(() -> {
                if (graph == null) {
                    reporterPane.setGraphReport(null);
                } else {
                    reporterPane.setGraphReport(GraphReportManager.getGraphReport(graph.getId()));
                }
            });
        }
    }

    @Override
    public void newPluginReport(final PluginReport pluginReport) {
        final Graph activeGraph = GraphManager.getDefault().getActiveGraph();
        if (activeGraph != null && pluginReport.getGraphReport().getGraphId().equals(activeGraph.getId())) {
            reporterPane.addPluginReport(pluginReport);
        }
    }

    @Override
    protected String createStyle() {
        return JavafxStyleManager.isDarkTheme()
                ? "resources/plugin-reporter-dark.css"
                : "resources/plugin-reporter-light.css";
    }

    @Override
    protected PluginReporterPane createContent() {
        return reporterPane;
    }
}
