/*
 * Copyright 2010-2024 Australian Signals Directorate
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package au.gov.asd.tac.constellation.utilities.datastructure;

import java.io.Serializable;
import java.util.Objects;
import java.util.stream.Stream;

/**
 * A data-structure for storing a triplet of related values.
 *
 * @author arcturus
 * @author cygnus_x-1
 *
 * @param <A> the type of the first object in the 4-tuple
 * @param <B> the type of the second object in the 4-tuple
 * @param <C> the type of the third object in the 4-tuple
 * @param <D> the type of the fourth object in the 4-tuple
 */
public class FourTuple<A, B, C, D> implements Serializable, Comparable<FourTuple<A, B, C, D>> {

    private static final long serialVersionUID = 1L;

    private A first;
    private B second;
    private C third;
    private D fourth;

    public FourTuple(final A first, final B second, final C third, final D fourth) {
        this.first = first;
        this.second = second;
        this.third = third;
        this.fourth = fourth;
    }

    /**
     * Convenience method for creating an appropriately typed tuple.
     *
     * @param <A> the type of the first object in the 4-tuple
     * @param <B> the type of the second object in the 4-tuple
     * @param <C> the type of the third object in the 4-tuple
     * @param <D> the type of the third object in the 4-tuple
     * @param first the first object in the tuple
     * @param second the second object in the tuple
     * @param third the third object in the tuple
     * @param fourth the fourth object in the tuple
     * @return a 4-tuple that is templated with the types of first, second,
     * third and fourth
     */
    public static <A, B, C, D> FourTuple<A, B, C, D> create(final A first, final B second, final C third, final D fourth) {
        return new FourTuple<>(first, second, third, fourth);
    }

    public A getFirst() {
        return first;
    }

    public void setFirst(final A first) {
        this.first = first;
    }

    public B getSecond() {
        return second;
    }

    public void setSecond(final B second) {
        this.second = second;
    }

    public C getThird() {
        return third;
    }

    public void setThird(final C third) {
        this.third = third;
    }

    public D getFourth() {
        return fourth;
    }

    public void setFourth(final D fourth) {
        this.fourth = fourth;
    }

    public Stream<Object> stream() {
        return Stream.of(first, second, third, fourth);
    }

    @Override
    public boolean equals(final Object obj) {
        if (obj == null || getClass() != obj.getClass()) {
            return false;
        }

        final FourTuple<?, ?, ?, ?> other = (FourTuple<?, ?, ?, ?>) obj;
        return Objects.equals(other.first, first)
                && Objects.equals(other.second, second)
                && Objects.equals(other.third, third)
                && Objects.equals(other.fourth, fourth);
    }

    @Override
    public int hashCode() {
        return (first == null ? 0 : first.hashCode())
                ^ (second == null ? 0 : second.hashCode())
                ^ (third == null ? 0 : third.hashCode())
                ^ (fourth == null ? 0 : fourth.hashCode());
    }

    @Override
    public String toString() {
        return String.format("(%s, %s, %s, %s)", first.toString(),
                second.toString(), third.toString(), fourth.toString());
    }

    /**
     * Compare the Tuple similar to in Python, by comparing the first followed
     * by second and finally the third. If the object is an instance of
     * {@link Comparable} then use the built in compareTo, else convert them to
     * a {@link String} and compare them.
     */
    @Override
    /**
     * We can suppress unchecked warnings for this method. They are generated by
     * the {Comparable} casts, however we manually check these are comparable
     * prior to the cast.
     */
    @SuppressWarnings("unchecked")
    public int compareTo(final FourTuple<A, B, C, D> o) {
        int compare;

        if (first instanceof Comparable compFirst && o.first instanceof Comparable compOFirst) {
            // compare the first using Comparable
            compare = compFirst.compareTo(compOFirst);
            if (compare != 0) {
                return compare;
            }
        } else {
            // compare the first using strings
            compare = first.toString().compareTo(o.first.toString());
            if (compare != 0) {
                return compare;
            }
        }

        if (second instanceof Comparable compSecond && o.second instanceof Comparable compOSecond) {
            // compare the second using Comparable
            compare = compSecond.compareTo(compOSecond);
            if (compare != 0) {
                return compare;
            }
        } else {
            // compare the second using strings
            compare = second.toString().compareTo(o.second.toString());
            if (compare != 0) {
                return compare;
            }
        }

        if (third instanceof Comparable compThird && o.third instanceof Comparable compOThird) {
            // compare the third using Comparable
            compare = compThird.compareTo(compOThird);
            if (compare != 0) {
                return compare;
            }
        } else {
            // compare the third using strings
            compare = third.toString().compareTo(o.third.toString());
            if (compare != 0) {
                return compare;
            }
        }

        if (fourth instanceof Comparable compFourth && o.fourth instanceof Comparable compOFourth) {
            // compare the fourth using Comparable
            compare = compFourth.compareTo(compOFourth);
            if (compare != 0) {
                return compare;
            }
        }

        // compare the fourth using strings
        return fourth.toString().compareTo(o.fourth.toString());
    }
}
